/* $Id: copybox.c,v 1.1 2001/10/13 11:01:44 thayne Exp $
******************************************************************************

   LibGGI linear 16 - copybox

   Copyright (C) 1997 Jason McMullan   [jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <string.h>

#include "linear-16-a8-z8-lib.h"


int
GGI_linear_16_a8_z8_copybox(ggi_visual *vis,
                            int x, int y,
                            int w, int h,
                            int nx, int ny)
{
	uint16 *src, *dest;
	int pixperline, stridenext;
	int line, linefirst, linelast, linenext;
	int pixel, pixelfirst, pixellast, pixelnext;

	LIBGGICLIP_COPYBOX(vis, x, y, w, h, nx, ny);
	PREPARE_FB(vis);

	pixperline = LIBGGI_FB_W_STRIDE(vis) / 2;

	if (ny < y) {
		linefirst = 0;
		linelast  = h;
		linenext  = 1;
		stridenext = pixperline;
	} else {
		linefirst = h - 1;
		linelast  = -1;
		linenext  = -1;
		stridenext = -pixperline;
	}

	if (ny < x) {
		pixelfirst = 0;
		pixellast  = w;
		pixelnext  = 1;
	} else {
		pixelfirst = w - 1;
		pixellast  = -1;
		pixelnext  = -1;
	}

	src  = (uint16 *)LIBGGI_CURREAD(vis)  + (y  + linefirst) * pixperline;
	dest = (uint16 *)LIBGGI_CURWRITE(vis) + (ny + linefirst) * pixperline;

#warning Im sure this isnt correct because it doesnt consider src alpha
#warning highly inefficient becasue I dont know how to access the A and Z buffs directly
	for (line = linefirst;
	     line != linelast;
	     line += linenext, src += stridenext, dest += stridenext) {
		for (pixel = pixelfirst; pixel != pixellast; pixel += pixelnext) {
			dest[pixel + nx] =
				calc_pixel(vis, pixel + nx, line + ny, src[pixel + x], dest[pixel + nx]);
		}
	}

	return 0;
}
