/*
******************************************************************************

   OVL implementation for X target - API functions.

   Copyright (C) 2001	Christoph Egger	[Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <math.h>

#include "xovl.h"

#include <ggi/display/xcommon.h>
#include <ggi/display/x.h>

#include <X11/X.h>

#include <ggi/ggi.h>



int OVL_X_tie(ggi_visual_t vis, ggiOvl_t ovl)
{
	int rc = GALLOC_OK;

	if (ovl == NULL) return GGI_EARGINVAL;
	if (ovl->res_handle_motor == NULL) return GGI_EARGINVAL;
	if (ovl->res_handle_carb == NULL) return GGI_EARGINVAL;
	if (ovl->res_handle_tank == NULL) return GGI_EARGINVAL;

	switch (ovl->res_type & GA_RT_TYPE_MASK) {
	case GA_RT_SPRITE:
		{
		struct xovl_Sprite *sprite = NULL;

		sprite = malloc(sizeof(struct xovl_Sprite));
		if (!sprite) return GGI_ENOMEM;
		ovl->Ovltype.sprite.target = (void *)sprite;

		sprite->sprite = ggiGAGetPriv(ovl->res_handle_tank);
		LIB_ASSERT(sprite->sprite != NULL);
		if (!sprite->sprite) {
			free(sprite);
			rc = GALLOC_EFAILED;
			goto err;
		}	/* if */

#warning Handle Motor-Grid here (div_min/max, mul_min/max).

		rc = GALLOC_OK;
		break;
		}

	case GA_RT_WINDOW:
		{
		struct xovl_Window *window = NULL;

		window = malloc(sizeof(struct xovl_Window));
		if (!window) return GGI_ENOMEM;
		ovl->Ovltype.window.target = (void *)window;

		window->window = ggiGAGetPriv(ovl->res_handle_tank);
		LIB_ASSERT(window->window != NULL);
		if (!window->window) {
			free(window);
			rc = GALLOC_EFAILED;
			goto err;
		}	/* if */

#warning Handle Motor-Grid here (div_min/max, mul_min/max).

		rc = GALLOC_OK;
		break;
		}

	case GA_RT_VIDEO:
		{
		struct xovl_Video *video = NULL;

		video = malloc(sizeof(struct xovl_Video));
		if (!video) return GGI_ENOMEM;
		ovl->Ovltype.video.target = (void *)video;

		video->video = ggiGAGetPriv(ovl->res_handle_tank);
		LIB_ASSERT(video->video != NULL);
		if (!video->video) {
			free(video);
			rc = GALLOC_EFAILED;
			goto err;
		}	/* if */

#warning Handle Motor-Grid here (div_min/max, mul_min/max).

		rc = GALLOC_OK;
		break;
		}
	}	/* switch */


	/* "Connect" motor and carb together

        Christoph: This should have been done before resources are checked/set
	Probably this call is a leftover from an older implementation?

	rc = ggiGAAttach2MotorGroup(reslist, ovl->res_handle_motor, 
	ovl->res_handle_carb);
	*/

	return GALLOC_OK;

err:
	return rc;
}	/* OVL_X_tie */



int OVL_X_setimage(ggiOvl_t ovl, ggi_visual_t vis,
		   long x, long y, unsigned char *alpha)
{
	ggi_xwin_common *priv = NULL;
	priv = LIBGGI_PRIVATE(ovl->vis);


	OVLDPRINT("%s:%s:%i: enter\n", DEBUG_INFO);


	switch (ovl->res_type & GA_RT_TYPE_MASK) {
	case GA_RT_SPRITE:

		switch (ovl->res_type & GA_RT_SUBTYPE_MASK) {
		case (GA_RT_SPRITE_DONTCARE & GA_RT_SUBTYPE_MASK):

			/* Oops! We know, X can handle kind of sprites:
			 * X-mouse-cursor-sprite and X-text-cursor-sprite.
			 * But nothing is specified to allocate, so we
			 * fail here.
			 */
			return GALLOC_EUNAVAILABLE;

		case (GA_RT_SPRITE_POINTER & GA_RT_SUBTYPE_MASK):
			{
			XColor fg, bg;
			ggi_color fg_col, bg_col;
			Pixmap source, mask;

			int xx, yy;

			struct xovl_Sprite *sprite;
			char *pix_buf = NULL;

			pix_buf = (char *)malloc(ovl->w * ovl->h *
						sizeof(ggi_pixel));
			if (!pix_buf) return GGI_ENOMEM;

			sprite = (struct xovl_Sprite *)
					ovl->Ovltype.sprite.target;

			OVLDPRINT("%s:%s:%i: sprite-size: (%i,%i)\n",
				DEBUG_INFO, ovl->w, ovl->h);

#if 0
			for (yy = 0; yy < ovl->h; yy++) {
				for (xx = 0; xx < ovl->w; xx++) {
					int i;
					ggi_color col;
					ggi_pixel pix;

					ggiGetPixel(vis, x+xx,y+yy, &pix);
					ggiUnmapPixel(vis, pix, &col);

				        /* NB: This formula is not very good...
					 */
				        pix = (col.r*82 +
						col.g*124 +
						col.b*50) / 256;

					pix = (int)rint((double)(pix / 65536.0f));
					LIB_ASSERT(pix == 0 || pix == 1);

					i = yy * ovl->w + xx;

					pix_buf[i] = pix;
				}	/* for */
			}	/* for */
#else
			ggiGetBox(vis, x,y, ovl->w,ovl->h, pix_buf);
#endif

			/* When this fails, then something
			 * is wrong in libGAlloc
			 */
			XFreeCursor(priv->x.display,
				sprite->sprite->cursor);

			fg.pixel = LIBGGI_GC_FGCOLOR(vis);
			bg.pixel = LIBGGI_GC_BGCOLOR(vis);

			ggiUnmapPixel(vis, fg.pixel, &fg_col);
			ggiUnmapPixel(vis, bg.pixel, &bg_col);

			fg.red = fg_col.r;
			fg.green = fg_col.g;
			fg.blue = fg_col.b;
			fg.flags = DoRed | DoGreen | DoBlue;

			bg.red = bg_col.r;
			bg.green = bg_col.g;
			bg.blue = bg_col.b;
			bg.flags = DoRed | DoGreen | DoBlue;

#if 1
			source = XCreatePixmapFromBitmapData(
					priv->x.display,
					priv->window,
					(signed char *)pix_buf,
					ovl->w, ovl->h,
					fg.pixel, bg.pixel,
					priv->visual.depth);
#else
			source = XCreateBitmapFromData(
					priv->x.display,
					priv->window,
					(signed char *)pix_buf,
					ovl->w, ovl->h);
#endif

			if (alpha == NULL) {
				mask = None;
			} else {
				mask = XCreateBitmapFromData(
					priv->x.display,
					priv->window,
					(signed char *)alpha,
					ovl->w, ovl->h);
			}	/* if */

			sprite->sprite->cursor = XCreatePixmapCursor(
				priv->x.display,
				source, mask,
				&fg, &bg, 0, 0);

			if (!ggiOvlIsHidden(ovl)) {
				XDefineCursor(priv->x.display,
					priv->window,
					sprite->sprite->cursor);
			}	/* if */


			if (mask != None) {
				XFreePixmap(priv->x.display, mask);
			}	/* if */

			XFreePixmap(priv->x.display, source);
			free(pix_buf);

			return GALLOC_OK;
			}

		case (GA_RT_SPRITE_CURSOR & GA_RT_SUBTYPE_MASK):
			{
			struct xovl_Sprite *sprite;

			sprite = (struct xovl_Sprite *)
					ovl->Ovltype.sprite.target;

			/* When this fails, then something
			 * is wrong in libGAlloc
			 */
			XFreeCursor(priv->x.display,
				sprite->sprite->cursor);

			/* Only Xlib has access to the font-structure
			 * in ggi_xlib_priv, so the X-target has a less
			 * flexible sprite-cursor.
			 */
			sprite->sprite->cursor = XCreateFontCursor(
					priv->x.display, CursorShape);

			if (!ggiOvlIsHidden(ovl)) {
				XDefineCursor(priv->x.display,
					priv->window,
					sprite->sprite->cursor);
			}	/* if */

			return GALLOC_OK;
			}

		case (GA_RT_SPRITE_SPRITE & GA_RT_SUBTYPE_MASK):
			/* true hw-sprites are not supported by X :-( */
			return GALLOC_EUNAVAILABLE;

		}	/* switch */

	case GA_RT_WINDOW:
		/* nothing todo here for this resource.
		 * The user takes care about its content.
		 */
		return GALLOC_OK;

	case GA_RT_VIDEO:

#warning FIXME: What should we do here?

		return GALLOC_EUNAVAILABLE;

	}	/* switch */

	return GALLOC_EFAILED;
}	/* OVL_X_setimage */





static int display(ggiOvl_t ovl)
{
	int rc = GALLOC_OK;

	switch (ovl->res_type & GA_RT_TYPE_MASK) {
	case GA_RT_SPRITE:

		rc = sprite_show(ovl);
		break;

	default:
		return GALLOC_EFAILED;
	}	/* switch */

	return rc;
}	/* display */






int OVL_X_setpos(ggiOvl_t ovl, long x, long y)
{
	int rc = GALLOC_OK;

	if (!ggiOvlIsEnabled(ovl)) {
		return GALLOC_EFAILED;
	}	/* if */

#warning Convert x,y to unit set in ovl->cb
#warning Handle snapping to grid here

	if (x == ovl->cur_x && y == ovl->cur_y) {
		return rc;	/* It's already there */
	}	/* if */

	ovl->cur_x = x;
	ovl->cur_y = y;

	if (ovl->hidecount != 0) {
		return rc;	/* It's hidden */
	}	/* if */

	_ggiOvlSync(ovl);


	rc = display(ovl);

	return rc;
}	/* OVL_X_setpos */



int OVL_X_move(ggiOvl_t ovl, long x, long y)
{
	int rc = GALLOC_OK;

	if (!ggiOvlIsEnabled(ovl)) {
		return GALLOC_EFAILED;
	}	/* if */

#warning Convert x,y to unit set in ovl->cb
#warning Handle snapping to grid here

	ovl->cur_x += x;
	ovl->cur_y += y;

	if (ovl->hidecount != 0) {
		return rc;	/* It's hidden */
	}	/* if */

	_ggiOvlSync(ovl);


	rc = display(ovl);

	return rc;
}	/* OVL_X_move */




int OVL_X_show(ggiOvl_t ovl)
{
	int rc = GALLOC_OK;

	if (ovl->hidecount > 0) {
		ovl->hidecount--;
	}	/* if */

	if (ovl->hidecount != 0) {
		return rc;	/* still hidden */
	}	/* if */


	rc = display(ovl);

	return rc;
}	/* OVL_X_show */



int OVL_X_hide(ggiOvl_t ovl)
{
	int rc = GALLOC_OK;

	if (ovl->hidecount == 0) {

		switch (ggiGA_TYPE(ovl)) {
		case GA_RT_SPRITE:

			rc = sprite_hide(ovl);
			break;

		default:
			break;
		}	/* switch */

	}	/* if */

	ovl->hidecount++;
	return rc;
}	/* OVL_X_hide */



int OVL_X_enable(ggiOvl_t ovl)
{
	int rc = GALLOC_EFAILED;

	if (ovl->disablecount > 0) {
		ovl->disablecount--;
	}	/* if */

	if (ovl->disablecount != 0) {
		return GALLOC_OK;	/* still disabled */
	}	/* if */


	switch (ggiGA_TYPE(ovl)) {
	case GA_RT_SPRITE:

#warning TODO
		break;

	case GA_RT_WINDOW:

#warning TODO
		break;

	case GA_RT_VIDEO:

#warning TODO
		break;

	default:
		return GALLOC_EFAILED;
	}	/* switch */

	return rc;
}	/* OVL_X_enable */



int OVL_X_disable(ggiOvl_t ovl)
{
	int rc = GALLOC_OK;

	if (ovl->disablecount == 0) {

		switch (ggiGA_TYPE(ovl)) {
		case GA_RT_SPRITE:
			break;

		case GA_RT_WINDOW:
			break;

		case GA_RT_VIDEO:
			break;

		default:
			return GALLOC_EFAILED;
		}	/* switch */

	}	/* if */


	ovl->disablecount++;	
	return rc;
}	/* OVL_X_disable */



int OVL_X_setsize(ggiOvl_t ovl, long w, long h)
{
	LIB_ASSERT(ovl != NULL);

#warning TODO

	/* fail as long as not ready implemented */
	return GALLOC_EFAILED;
}	/* OVL_X_setsize */


int OVL_X_setpriority(ggiOvl_t ovl, uint32 slot, int swap)
{
	LIB_ASSERT(ovl != NULL);

#warning TODO

	/* fail as long as not ready implemented */
	return GALLOC_EFAILED;
}	/* OVL_X_setpriority */


int OVL_X_setflags(ggiOvl_t ovl, enum ggiOvl_flags flags)
{
	LIB_ASSERT(ovl != NULL);

#warning TODO

	/* fail as long as not ready implemented */
	return GALLOC_EFAILED;
}	/* OVL_X_setflags */


int OVL_X_setzoom(ggiOvl_t ovl, long x, long y, long w, long h, int keep_ratio)
{
	LIB_ASSERT(ovl != NULL);


#warning Convert x,y,w and h to unit set in ovl->cb

	if (keep_ratio) {
		double ratio;

		if (w < h) {
			ratio = ovl->cur_h / ovl->cur_w;
			h = w * ratio;
		} else {
			ratio = ovl->cur_w / ovl->cur_h;
			w = h * ratio;
		}	/* if */		
	}	/* if */


#warning TODO

	/* fail as long as not ready implemented */
	return GALLOC_EFAILED;
}	/* OVL_X_setzoom */
